package com.prlab.idpserver.controller;

import com.prlab.idpserver.hkdf.HKDF;
import com.prlab.idpserver.model.IdentityRequestV2;
import org.apache.tomcat.util.buf.HexUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.util.Base64Utils;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RestController;

import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;
import java.math.BigInteger;
import java.nio.ByteBuffer;
import java.nio.charset.StandardCharsets;
import java.security.InvalidKeyException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.sql.Types;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicReference;

@RestController
@RequestMapping("/api")
public class askIdentityV2 {

	@Autowired
	JdbcTemplate jdbcTemplate = new JdbcTemplate();
	Logger logger = LoggerFactory.getLogger(askIdentityV2.class);


	@PostMapping("/idp/askIdentityV2")
	public String queryIdentity(@RequestBody IdentityRequestV2 r){
		logger.info("timestamp askIdentityV2 {}", System.currentTimeMillis());

		logger.info("r : {}, {}, {}, {}", r.nonce_base64, r.idHash_base64, r.cE_base64, r.HMAC_base64);
		byte[] nonce = Base64Utils.decodeFromString(r.nonce_base64);
		byte[] idHash = Base64Utils.decodeFromString(r.idHash_base64);
		byte[] cE = Base64Utils.decodeFromString(r.cE_base64);
		byte[] hmac = Base64Utils.decodeFromString(r.HMAC_base64);
		logger.info("nonce: {}\nidHash: {}\ncE: {}\nhmac: {}\n",
				HexUtils.toHexString(nonce),
				HexUtils.toHexString(idHash),
				HexUtils.toHexString(cE),
				HexUtils.toHexString(hmac)
				);


		String sql_query_by_id_hash ="select * from pivinfo " +
				"where hex(?)=sha2(unhex(concat(sha2(id, 256), sha2(?, 256))), 256) limit 1;" ;

//		String sql_query_by_id_hash = "SELECT * FROM users " +
//				"WHERE hex(?)=sha2(unhex(concat(sha2(uid, 256), sha2(?, 256))), 256)";

		long t1 = System.currentTimeMillis();
		List<Map<String, Object>> list =
				jdbcTemplate.queryForList(sql_query_by_id_hash,
						new Object[]{
								Base64Utils.decodeFromString(r.idHash_base64),
								Base64Utils.decodeFromString(r.nonce_base64)},
						new int[]{Types.VARBINARY, Types.VARBINARY});
		long t2 = System.currentTimeMillis();
		logger.info("timestamp dbt1 {}", t1);
		logger.info("timestamp dbt2 {}", t2);
		logger.info("time consumed: {} ms", t2-t1);


		BigInteger nonceInt = new BigInteger(nonce);
		BigInteger nonceInt1 = nonceInt.add(new BigInteger(String.valueOf(1)));
		AtomicReference<String> msg = new AtomicReference<>("Your Identity is Invalid.," + (t2 - t1));

		list.forEach(map->{
			String id = (String)map.get("id");
			try {
				MessageDigest sha256 = MessageDigest.getInstance("SHA-256");
				byte[] idSha256 = sha256.digest(id.getBytes(StandardCharsets.UTF_8));
				byte[] nonce1Sha256 = sha256.digest(nonceInt1.toByteArray());
				sha256.update(idSha256);
				byte[] idHash1 = sha256.digest(nonce1Sha256);
				byte[] counter = new byte[cE.length];
				xor(cE, idHash1, counter);
				ByteBuffer byteBuffer = ByteBuffer.allocate(Long.BYTES); // uint32 to int64(long)
				byteBuffer.put(new byte[4]);
				byteBuffer.put(counter);
				byteBuffer.rewind();
				long c = byteBuffer.getLong();
				logger.info("id: {}",id);
				logger.info("c: {}",c);
				HKDF hkdf = new HKDF((byte[])map.get("sharedSecret"), (byte[])map.get("hkdf_chain_iv"), c);
				long t3 = System.nanoTime();
				byte[] outputKey = hkdf.getOutputKey(((String)map.get("id")).getBytes(StandardCharsets.UTF_8), 32);
				long t4 = System.nanoTime();
				logger.info("t4-t3: {}", t4-t3);
				Mac h256 = Mac.getInstance("HmacSHA256");
				h256.init(new SecretKeySpec(outputKey, "HmacSHA256"));
				byte[] hmac_ = h256.doFinal(idHash);
				print("hmac_", hmac_);
				print("hmac", hmac);
				if(Arrays.equals(hmac, hmac_)){
					logger.info("hmac is correct");
					msg.set("Your Identity is Valid.," + (t2 - t1));
				}
			} catch (NoSuchAlgorithmException | InvalidKeyException e) {
				throw new RuntimeException(e);
			}
		});
		logger.info("timestamp askIdentityV2return {}", System.currentTimeMillis());
		if(list.size()>0)
//			return (String)list.get(0).get("id");
			return msg.get();
		else
			return "IDENTITY_NOT_FOUND";
	}

	public void xor(byte[] a, byte[] b, byte[] output){
		for(short i = 0; i< a.length; i++){
			output[i] = (byte)(a[i] ^ b[i]);
		}
	}

	public void print(byte[] data){
		logger.info(HexUtils.toHexString(data));
	}
	public void print(String tag, byte[] data){
		logger.info("{}: {}", tag, HexUtils.toHexString(data));
	}
}
