package com.prlab.idpserver.hkdf;

import javax.crypto.SecretKey;
import java.nio.ByteBuffer;
import java.security.NoSuchAlgorithmException;
public class HKDF {


    private final byte[] key; 
    private final byte[] iv; 
    private long iterate;
    private HMAC hmac;

    public HKDF(byte[] key, byte[] iv, long iterate ) throws NoSuchAlgorithmException {
        this.key = key;
        this.iv = iv;
        this.iterate = iterate;
        this.hmac = new HMAC(key);
    }
    private byte[] extract(byte[] salt, byte[] ikm) throws NoSuchAlgorithmException {
        hmac.setKey(salt);
        hmac.update(ikm);
        return hmac.doFinal();
    }
    private byte[] expand(byte[] prk, byte[] info, int outLengthBytes) throws NoSuchAlgorithmException{
        // sha256 maclength is always 32
        int n = (int) Math.ceil(((double) outLengthBytes) / ((double) 32));
        if( n > 255) {
           throw new IllegalArgumentException("out length must be maximal 255 * hash-length; requested: " + outLengthBytes + " bytes");      
        }
        byte[] blockN;
        byte[] ret = new byte[outLengthBytes];
        int stepSize = 0;
        int remainingBytes = outLengthBytes;
        for(int i = 0; i < n; i++ ){
            hmac.setKey(prk);
            hmac.update(info);
            byte[] counter = {(byte)(i+1)};
            hmac.update(counter);
            blockN = hmac.doFinal();

            stepSize = blockN.length;
            if(stepSize < remainingBytes){
                arrayCopy(ret, blockN, outLengthBytes-remainingBytes, stepSize);
                remainingBytes -= stepSize;
            }else{
                arrayCopy(ret, blockN, outLengthBytes - remainingBytes, remainingBytes);
            }
        }
        return ret;
        
    }

    public byte[] getPRK() throws NoSuchAlgorithmException{
        byte[] prf = this.iv;        
        for( long i=0 ; i < this.iterate; i++ ){
            prf = extract(prf, this.key);
        }
        return prf;
    }
    public byte[] getOutputKey(byte[] info, int outLengthBytes) throws NoSuchAlgorithmException{
        byte[] outputKey;
        byte[] salt = this.iv;
        for( long i = 0; i < this.iterate; i++){
            salt = extract(salt, this.key);
        }
        outputKey = expand(salt, info, outLengthBytes);
        return outputKey;
    }
    private static void arrayCopy(byte[] arr1, byte[] arr2, int start, int offset){
        for(int i = 0; i < offset; i++ ){
            arr1[start+i] = arr2[i];
        }
    }
}
